/* *****************************************************
 * cropkoala - crops an given Koala image to a smaller image cut out by a given window
 *
 * September-2020 V0.3
 * Wilfried Elmenreich
 * Code under CC-BY-4.0 license
 * **************************************************** */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>


typedef unsigned char int8_t;

typedef struct { 
 int8_t startlo;
 int8_t starthi;
 int8_t data[0x10000];
 int Bstart;
 int Sstart;
 int Cstart;
 int Gstart; 
 int Bend;
 int Send;
 int Cend;
 int length;
} cbmfile_t;

int VERBOSE=1;
char *progname;
char *palette="17df3a5ce842b960";
char *fstructure="BSCG";

cbmfile_t load_file(char* filename) {
  FILE *fp;
  cbmfile_t filedata;
  fp = fopen(filename, "rb");
  filedata.startlo = fgetc(fp);
  filedata.starthi = fgetc(fp);
  filedata.length=0;
  while(1) {
    int c=fgetc(fp);
    if (feof(fp)) break;  
    filedata.data[filedata.length++]=c;
  }
  fclose(fp);
  if (VERBOSE) {
      int start=filedata.startlo+256*filedata.starthi;
      int end=start+filedata.length;
      printf("Read file '%s' $%04x-$%04x\n",filename,start,end);
  }
  if (filedata.length!=10001) //if the size fits, we assume this is a Koala image
	  printf("Warning! File length is not 10001 bytes but %i.\n",filedata.length);
  filedata.Bstart=0;
  filedata.Bend=filedata.Sstart=8000;
  filedata.Send=filedata.Cstart=9000;
  filedata.Cend=filedata.Gstart=10000;     	
  
return filedata;
}
 
void save_file(cbmfile_t filedata, char* filename) {
   FILE *fp;
   fp = fopen(filename, "wb");
   fprintf(fp, "%c", filedata.startlo);
   fprintf(fp, "%c", filedata.starthi);
   for (int i=0; i<filedata.length; i++)
      fprintf(fp, "%c", filedata.data[i]);
   fclose(fp);
   if (VERBOSE) {
      int start=filedata.startlo+256*filedata.starthi;
      int end=start+filedata.length;
      printf("Written file '%s' $%04x-$%04x\n",filename,start,end);
   }
}

cbmfile_t crop_koala(cbmfile_t filedata,int x0p,int y0p,int widthp,int heightp) {
  cbmfile_t croppedfiledata;

  croppedfiledata.startlo = filedata.startlo;
  croppedfiledata.starthi = filedata.starthi;  
  croppedfiledata.length = 0;
  
  int x0=x0p/8;
  int y0=y0p/8;
  int width=(x0p+widthp+7)/8 - x0;
  int height=(y0p+heightp+7)/8 - y0;
  
  const int xmasksleft[8]={ 0xff, 0x7f, 0x3f, 0x1f, 0xf, 0x7, 0x3, 0x1 };
  const int xmasksright[8]={ 0xff, 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe };
  int xmask0=xmasksleft[x0p & 7];
  int xmask1=xmasksright[(x0p+widthp) & 7];
  
  printf("x0 %i y0 %i xmask0 %x xmask1 %x",x0,y0,xmask0,xmask1);

  // copy relevant parts of bitmap
  for (int y=y0;y<y0+height;y++)  
    for (int x=x0;x<x0+width;x++) {
      int addr=filedata.Bstart + y*320 + x*8;
      int i0=0;
      int i1=7;
      if (y==y0) i0=y0 & 7;
      if (y==y0+height-1) i1=(y0p+heightp) & 7;
      for(int i=0;i<8;i++) {
      	int data=filedata.data[addr+i]; 
      	if (x==x0) data = data & xmask0;
      	if (x==x0+width-1) data = data & xmask1;      	
      	croppedfiledata.data[croppedfiledata.length++]=data;      	
	  }    	
    }
  croppedfiledata.Bend=croppedfiledata.Sstart=croppedfiledata.length;

  // copy relevant parts of screen 
  for (int y=y0;y<y0+height;y++)  
    for (int x=x0;x<x0+width;x++) {
      int addr=filedata.Sstart + y*40 + x;
      croppedfiledata.data[croppedfiledata.length++]=filedata.data[addr];
    }
  croppedfiledata.Send=croppedfiledata.Cstart=croppedfiledata.length;

  // copy relevant parts of colram 
  for (int y=y0;y<y0+height;y++)  
    for (int x=x0;x<x0+width;x++) {
      int addr=filedata.Cstart + y*40 + x;
      croppedfiledata.data[croppedfiledata.length++]=filedata.data[addr];
    }
  croppedfiledata.Cend=croppedfiledata.Gstart=croppedfiledata.length;
  
  //we omit the background value
  return croppedfiledata;
}


void usage(char *s) {
   if (s!=NULL) fprintf(stderr,"%s\n\n",s);
	
   fprintf(stderr,"Usage: %s [-i <infilename>] [-o <outfilename>][-x X][-y Y][-w width][-h height]\n",progname);
   fprintf(stderr," -i file in Koala format to be read\n");     
   fprintf(stderr," -o File to be written\n"); 
   fprintf(stderr," -x starting X coordinate from 0..319\n");  
   fprintf(stderr," -y starting Y coordinate from 0..199\n");
   fprintf(stderr," -w width in pixels\n");     
   fprintf(stderr," -h heigh in pixels\n");     
   fprintf(stderr,"\nx coordinates should be always a multiple of 2\n");   
           
   exit(1);
}

int main(int argc, char ** argv) {
  int c;
	
  char *infilename=NULL;
  char *outfilename=NULL;
  
  int x0=0,y0=0,width=40,height=25;

  progname=argv[0];
  while ((c=getopt(argc,argv,"i:o:x:y:w:h:"))!=EOF)
  {
     switch(c)
     {
     case 'i':
     	if (infilename!=NULL) usage("Only one infile can be given.");
        infilename=optarg;
        break;      	
     case 'o':
     	if (outfilename!=NULL) usage("Only one outfile can be given.");     	
        outfilename=optarg;
        break;       
     case 'x':
        x0=atol(optarg);
        break;          		           
     case 'y':
        y0=atol(optarg);
        break;      
     case 'w':
        width=atol(optarg);
        break;          		           
     case 'h':
        height=atol(optarg);
        break;                       
     default:
        usage(NULL);
     }
  }
  
  if (optind-argc!=0) usage(NULL);
  
  if (infilename==NULL) usage("No infile given!");
  if (outfilename==NULL) usage("No outfile given!");

  cbmfile_t filedata=load_file(infilename);

  cbmfile_t croppedfiledata=crop_koala(filedata,x0,y0,width,height);

  save_file(croppedfiledata,outfilename);

  return 0;
}


